import interlang;
import std.string;
import core.stdc.string: strlen;
import std.conv;
import std.stdio;
import robject;

auto cstr2dstr(inout(char)* cstr){
	return cast(string) (cstr ? cstr[0 .. strlen(cstr)] : "");
}

/**
 * Operand Code listesi.
 * ubyte olarak tanımladığımız için max 255 tane operand tanımlayabilirsiniz.
*/
enum il : ubyte{
	hlt,
	load, loadvar, definevar,
	push, pushparam, call,
	add, sub, div, mul,
}

/**
 * Nesnelere ulaşılamaması durumunda onları hayatta tutmak için
 * aşağıdaki version kodunu aktif edin.
*/
//version = save_objects;
class IL{
	ubyte[] codes;///IL kodlarımızın saklandığı dizi.
	version(save_objects) RObject[] objects;///Olası nesne silinmesini engellemek için nesne dizimiz.
	//Nesne yükleme fonksiyonumuz
	void load(RObject addr){
		version(save_objects) objects ~= addr;
		codes ~= il.load;
		codes ~= (cast(ubyte*) &addr)[0..(void*).sizeof];
	}
	/* load fonksiyonları */
	void load(int val){
		load(new RNumber(val));
	}
	void load(bool val){
		load(new RBoolean(val));
	}
	void load(string val){
		load(new RString(val));
	}
	/* Değişken yükle */
	void loadvar(string name){
		codes ~= il.loadvar;
		codes ~= cast(ubyte[]) name;
		codes ~= 0x00;///Bitiş karakteri
	}
	/* Değişken tanımlama */
	void definevar(string name){
		codes ~= il.definevar;
		codes ~= cast(ubyte[]) name;
		codes ~= 0x00;
	}
	/* Fonksiyon Çağır */
	void call(size_t parameter_count){
		codes ~= il.call;
		codes ~= (cast(ubyte*) &parameter_count)[0..size_t.sizeof];
	}
	/* Stacke parametre yaz */
	void pushparam(){
		codes ~= il.pushparam;
	}
	/* Stacke yaz */
	void push(){
		codes ~= il.push;
	}
	/* Toplama */
	void add(){
		codes ~= il.add;
	}
	/* Çıkarma */
	void sub(){
		codes ~= il.sub;
	}
	/* Bölme */
	void div(){
		codes ~= il.div;
	}
	/* Çarpöa */
	void mul(){
		codes ~= il.mul;
	}
	/* Makineyi sonlandır */
	void hlt(){
		codes ~= il.hlt;
	}
	/* IL kodu yaz */
	void newcode(il code){
		codes ~= code;
	}
	/* Assembly tarzı IR çıktısı verir. */
	@property override string toString(){
		string output;
		auto ptr = codes.ptr;
		start:
		switch(*cast(il*) ptr){
			case il.loadvar: ptr++;
				string name = (cast(char*) ptr).cstr2dstr();
				output ~= "loadvar %s\n".format(name);
				ptr += name.length + 1;
				goto start;
			case il.definevar: ptr++;
				string name = (cast(char*) ptr).cstr2dstr();
				output ~= "definevar %s\n".format(name);
				ptr += name.length + 1;
				goto start;
			case il.push, il.pushparam, il.sub, il.add, il.div, il.mul:
				output ~= "%s\n".format(*cast(il*) ptr);
				ptr++;
				goto start;
			case il.load: ptr++;
				RObject addr = *cast(RObject*) ptr;
				output ~= "load %s\n".format(addr);
				ptr += (void*).sizeof;
				goto start;
			case il.call: ptr++;
				size_t addr = *cast(size_t*) ptr;
				output ~= "call %s\n".format(addr);
				ptr += size_t.sizeof;
				goto start;
			case il.hlt: break;
			default:
				throw new Exception("Bilinmeyen operand code %s".format(*cast(il*) ptr));
		}
		return output;
	}
}