module parser;
import std.string;
import lexer;
import interlang;
import std.conv : to;

/// Fonksiyonlar arası iletişimi sağlayan dönüş değerini belirten yapı.
enum RetType{
	none, expression
}

/// İşlem önceliği tablosuna yeniden erişmemek için öncelik değerini tuttuğumuz yapı.
struct Operator{
	Type type; /// Operatör
	size_t precedence; ///İşlem önceliği
}

/**
 Belirlediğimiz gramer kurallarına göre kodu işleyecek olan sınıfımız.
 */
class Parser : Lexer{
	private:
		Token[] tokens;
		IL il;
		size_t i;// Şu anki token
		static size_t[Type] op_precedences;/// İşlem Önceliği tablosu
		static interlang.il[Type] op_ilcodes;/// Operatörlerin IL Code karşılığı
		static this(){
			/**
				İşlem önceliği yüksek olan daha büyük sayı ile gösteriliyor.
			*/
			op_precedences = [Type.plus: 2 , Type.minus: 2, Type.times: 3, Type.divide: 3];
			op_ilcodes = [Type.plus: interlang.il.add , Type.minus: interlang.il.sub, Type.times: interlang.il.mul, Type.divide: interlang.il.div];
		}

	public:
		/// Kodlarımızı lexera gönderecek ve sonra sözdizimsel analizi başlatacak olan ana fonksiyonumuz.
		IL parse(string code){
			il = new IL();
			tokens = lexit(code);/// Lexerdan gelen tokenleri tokens dizisine yükle.
 			while(i < tokens.length){/// Token olduğu sürece
				switch(tokens[i].type){
					case Type.identifier:/// Bir tanımlayıcı yani değişken gelirse
						auto name = tokens[i].value;/// Sonraki token'e geçileceği için ismini değişkenin adını saklıyoruz.
						i++;/// index değerini arttırarak sonraki tokene geçiyoruz.
						size_t pcount;
						if(tokens[i].type==Type.equals){///Bir tanımlama ise
							i++;/// = karakterini geçiyoruz.
							if(calcIt() == RetType.expression){
								il.definevar(name);
							}else throw new Exception("= den sonra bir ifade bekleniyordu. %s geldi".format(tokens[i].type));
						}else if(getFunction(pcount)){ /// Bir fonksiyon çağırma ise
							il.loadvar(name);
							il.call(pcount);
						}else{
							throw new Exception("%s değişkeni üzerinde hiçbir işlem yapılmadı".format(name));
						}
						break;
					default:/// Switch içerisindeki hiçbir yerle eşleşemezse hata fırlatıyoruz.
						throw new Exception("Beklenmeyen tür: %s".format(tokens[i].type));
				}
				if(tokens[i].type==Type.semicolon){/// Noktalı virgül kullanımını zorunlu hale getir
					i++;
					continue;
				}else throw new Exception("; bekleniyordu %s geldi".format(tokens[i].type));
			}
			il.hlt();
			return il;
		}
	private:
		RetType getIt(){
			switch(tokens[i].type){
				case Type.number:
					il.load(to!int(tokens[i].value));
					i++;
					return RetType.expression;
				case Type.string:
					il.load(tokens[i].value);
					i++;
					return RetType.expression;
				case Type._true:
					il.load(true);
					i++;
					return RetType.expression;
				case Type._false:
					il.load(false);
					i++;
					return RetType.expression;
				case Type.identifier:
					string name = tokens[i].value;
					i++;
					size_t pcount;
					if(getFunction(pcount)){
						il.loadvar(name);
						il.call(pcount);
 					}else il.loadvar(name);
					return RetType.expression;
				default:
					return RetType.none;
			}
		}

		/**
			Shunting-Yard algoritmasını kullanarak matematiksel ifadeleri işlem önceliğine
			göre ayrıştıran fonksiyon.
		*/
		RetType calcIt(){
			Operator[] operators; /// Operatörlerin geçici olarak saklandığı operatör stack
			bool waitexp = true; /// İşleçden sonra matematiksel ifade gelip gelmediğini kontrol etmek için
			while(i < tokens.length){
				if(getIt() == RetType.expression) waitexp = false; /// İlk ifademizi okuyoruz.
				else if(waitexp) break; /// ilk if koşulumuzda ifade gelmedi ise ve bir ifade bekleniyorsa döngüyü kır.
				else if(auto p = tokens[i].type in op_precedences){/// İşlem önceliğimizi öncelik tablosundan al
					while(operators.length > 0 && *p < operators[$-1].precedence ){/// Eğer operatör stackda eleman var ise ve işlem önceliğimiz daha küçük ise
						il.newcode(op_ilcodes[operators[$-1].type]);
						operators = operators[0..$-1];/// Operatörü stackdan sil.
					}
					il.push();
					operators ~= Operator(tokens[i].type, *p);///Stack'e operatörü ekle
					i++;
					waitexp = true;/// Artık bir işleç geldiği için bir ifade gelmesi gerekecek.
				}else break;
			}
			if(waitexp) throw new Exception("Bir ifade bekleniyordu! %s geldi".format(tokens[i].type));
			foreach_reverse(operator; operators){/// Operatör stackda kalan bütün operatörleri işle.
				il.newcode(op_ilcodes[operator.type]);
			}
			return RetType.expression;
		}
		
		/**
			Fonksiyon çağırma işlemlerini yakalar.
		*/
		bool getFunction(out size_t pcount){
			if(tokens[i].type==Type.lparen) i++;/// Eğer ( tokeni var ise sonraki tokene geçer yoksa fonksiyondan çıkar.
			else return false;
			if(tokens[i].type==Type.rparen) goto end; /// Parantez kapatıldı ise döngüye girmeden direk sona atlar.
			while(i < tokens.length){
				if(calcIt() == RetType.none){
					throw new Exception("Bir ifade beklenirken %s geldi".format(tokens[i].type));
				}else{
					pcount++;
					il.pushparam();
				}
				if(tokens[i].type == Type.comma){ /// Sonraki parametre için 'virgül' kontrolü.
					i++;
					continue;
				}else break;
			}
			if(i >= tokens.length) throw new Exception("Parantezin kapatılması bekleniyordu.");
			else if(tokens[i].type!=Type.rparen) throw new Exception("Parantezin kapatılması bekleniyordu. %s geldi".format(tokens[i].type));
		end:
			i++;
			return true;
		}
}